/*
 * For reading bits on general purpose I/O (gpio) ports.
 *
 *
 * (C) Copyright 2001-2007, Axis Communications AB, LUND, SWEDEN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <syslog.h>
#include "gpio.h"

char gpio_port[] = GPIO_PORT;
unsigned char gpio_bit_count[] = GPIO_BIT_COUNT;
int read_outbits = 0;

const unsigned int MAX_PORTS = sizeof(gpio_port);

void usage(void) 
{
	unsigned int i;
	char *s = "Usage:";
	for (i = 0; i < MAX_PORTS; i++) {
		fprintf(stderr, "%s readbits [-o] [ -p %c [ -b 0-%d ] ]\n",
		s, gpio_port[i], gpio_bit_count[i] - 1);
		s = "   or:";
        }
	fprintf(stderr, "%s readbits [-o] [ -a <port><bit> ]\n", s);
	exit(1);
}

int get_port_index(char port)
{
	unsigned int i;
	int idx = -1;

	for (i = 0; i < MAX_PORTS; i++) {
		if (gpio_port[i] == port) {
			idx = i;
			break;
		}
        }

	return idx;
}

void print_bits(int fd, int bit, int maxBits) 
{
	unsigned long dir=0;
	unsigned long data=0;
	if (read_outbits) {
		ioctl(fd, _IO(GPIO_IOCTYPE, IO_SETGET_OUTPUT), &dir);
		dir = ~dir;
		ioctl(fd, _IO(GPIO_IOCTYPE, IO_READ_OUTBITS), &data);
	} else {
		ioctl(fd, _IO(GPIO_IOCTYPE, IO_SETGET_INPUT), &dir);
		dir = ~dir;
		ioctl(fd, _IO(GPIO_IOCTYPE, IO_READ_INBITS), &data);
	}

	if (bit == -1) {
		int i;

		for (i = maxBits-1; i >= 0; i--) {
			printf("%c", (dir & (1 << i))? 'X' :
			       (data & (1 << i))? '1' : '0');
		}
	} else {
		printf("%c", (dir & (1 << bit))? 'X' :
		       (data & (1 << bit))? '1' : '0');
	}
}



int main(int argc, char **argv)
{
	int opt;
	int port = -1;
	int port_index = -1;
	int bit = -1;
	int fd;
	int startport, endport;
	char fname[16]; /* Room for /dev/gpiox */
	

	openlog(argv[0], LOG_PID, LOG_USER);
	
	while ((opt = getopt(argc, argv, "ob:p:a:")) != -1) {
		switch(opt) {
			case 'o':
				read_outbits = 1;
				break;
			case 'b':
			{
				char* endptr;
				bit = strtol(optarg, &endptr, 10);
				if ((endptr == optarg) || (*endptr != '\0')) {
					usage();
				}
				break;
			}
			case 'p':
				if (strlen(optarg) == 1) {
					port = tolower(*optarg);

					if ((port_index = get_port_index(port)) < 0) {
						usage();
					}
				} else {
					usage();
				}
				break;
			case 'a':
				if (strlen(optarg) > 1) {
					port = tolower(*optarg);

					if ((port_index = get_port_index(port)) < 0) {
						usage();
					} else {
						char* endptr;
						bit = strtol(optarg+1, &endptr, 10);
						if ((endptr == optarg+1) || (*endptr != '\0')) {
							usage();
						}						
					}
				} else {
					usage();
				}
				break;
			default:
				usage();
		}
	}

	if ((port_index == -1) && (bit != -1)) {
		usage();
	}
	
	if (port_index == -1) {
		startport = 0;
		endport = MAX_PORTS - 1;
	} else {
		if (bit < -1 || bit > (gpio_bit_count[port_index] - 1)) {
			usage();
		}
		startport = endport = port_index;
	}

	for (port = startport; port <= endport; port++) {
			sprintf(fname, "/dev/gpio%c", gpio_port[port]);
			if ((fd = open(fname, O_RDWR)) < 0) {
				int err_saved = errno;
				perror("open");
				syslog(LOG_ERR, "Unable to open device %s. %s", fname, strerror(err_saved));
				exit(1);
			} else {
				if (port != startport)
					printf(" ");
				print_bits(fd, bit, gpio_bit_count[port]);
				close(fd);
			}
	}

	printf("\n");

	closelog();

	return 0;
}
